// hlavickove subory *******************************************************
#include "../includes.h"
// vlastny .h subor
#include "phwall.h"

// mur ---------------------------------------------------------------------

// konstruktor *************************************************************
CPhWall::CPhWall()
{
 t_scale=1.0f;       // zrychlenie/spomalenie plynutia casu
 t_step=1.0f/100.0f; // trvanie 1 vypoctu simulacie
 // cas, ktory uplynul od posledneho zaznamenaneho casoveho okamihu
 t_next=0.0f;
 // casovy okamih po vykonani x vypoctov simulacie
 t_last=0.0f;
 // pocet opakovani vypoctov simulacie
 iter_num=1;

 // pomocna matica pre nastavenie tenzorov zotrvacnosti
 mat3x3 I;

 // tehly, z ktorych je zlozeny mur
 bricks_num=10; // pocet tehal, z ktorych je zlozeny mur
 loopi(bricks_num) {
  // polovica rozmerov RB
  bricks[i].halfSize=vec3(0.4f,0.3f,0.4f);
  // inicializacia RB
  bricks[i].rb->rb_init(bricks[i].halfSize);
  // nastavenie parametrov RB
  bricks[i].rb->set_M(100.0f);
  I.set_box_inertia_tensor(bricks[i].halfSize,100.0f);
  bricks[i].rb->set_Ibs(I);
  // normalizacia R, vypocet t_mat, I_inv - vzdy po zmene R, alebo Ibs
  bricks[i].rb->update_params();
  // poloha RB
  switch (i) {
   case 0: bricks[i].rb->set_s(-8.0f,-1.1f,9.2f); break;
   case 1: bricks[i].rb->set_s(-8.0f,-1.1f,8.6f); break;
   case 2: bricks[i].rb->set_s(-8.0f,-1.1f,8.0f); break;
   case 3: bricks[i].rb->set_s(-8.0f,-1.1f,7.4f); break;
   case 4: bricks[i].rb->set_s(-8.0f,-1.1f,6.8f); break;
   case 5: bricks[i].rb->set_s(-8.0f,-0.5f,9.2f); break;
   case 6: bricks[i].rb->set_s(-8.0f,-0.5f,8.6f); break;
   case 7: bricks[i].rb->set_s(-8.0f,-0.5f,8.0f); break;
   case 8: bricks[i].rb->set_s(-8.0f,-0.5f,7.4f); break;
   case 9: bricks[i].rb->set_s(-8.0f,-0.5f,6.8f); break;
   default: break;
  }
  // pociatocne gravitacne zrychlenie zapricinene gravitacnou silou
  bricks[i].rb->set_a(vec3(0.0f,-9.81f,0.0f));
  bricks[i].rb->set_damping(0.9f,0.9f);
  bricks[i].rb->set_friction(0.9f);
 }

 // kolizne roviny
 planes[0].normal=vec3(0.0f,1.0f,0.0f);
 planes[0].dist=1.4f;
 planes[1].normal=vec3(1.0f,0.0f,0.0f);
 planes[1].dist=10.0f;

 // nahratie textury pre 10 tehal
 texture.load2D(tid[100],"data/textures/brick.png",T_LINEAR|T_CLAMP);
 texture.load2D(tid[101],"data/textures/brick.png",T_LINEAR|T_CLAMP);
 texture.load2D(tid[102],"data/textures/brick.png",T_LINEAR|T_CLAMP);
 texture.load2D(tid[103],"data/textures/brick.png",T_LINEAR|T_CLAMP);
 texture.load2D(tid[104],"data/textures/brick.png",T_LINEAR|T_CLAMP);
 texture.load2D(tid[105],"data/textures/brick.png",T_LINEAR|T_CLAMP);
 texture.load2D(tid[106],"data/textures/brick.png",T_LINEAR|T_CLAMP);
 texture.load2D(tid[107],"data/textures/brick.png",T_LINEAR|T_CLAMP);
 texture.load2D(tid[108],"data/textures/brick.png",T_LINEAR|T_CLAMP);
 texture.load2D(tid[109],"data/textures/brick.png",T_LINEAR|T_CLAMP);

 // vyhradenie pamate pre obrazok diery po gulke
 data=new unsigned char[8*4*3];
 loopi(8*4*3) data[i]=0;
}

// destruktor **************************************************************
CPhWall::~CPhWall()
{
 DELETEA(data);
}

// vykreslenie muru ********************************************************
void CPhWall::draw(vec3 EP,vec3 TP,unsigned char F,bool CF,float DT)
{
 // nebolo zasiahnute ziadne RB
 hit_RB=-1;

 // ak Avatar striela
 if ((F!=0)&&(CF)) {
  // normalizovany dopredu smerujuci vektor
  f_vec=-TP;
  // vzdialenost, minimalna vzdialenost
  d=0.0f; d_min=wRange;
  // vrchol na pozicii maximalneho dostrelu
  TP=EP+TP*wRange;
  // suradnice Avatara su opacne ako suradnice v scene
  EP=-EP; TP=-TP;

  // najdenie najblizsieho trojuholnika patriaceho RB, ktory bol zasiahnuty gulkou
  // pre vsetky RB
  loopi(bricks_num) {
   // zaciatok a koniec luca zacinajuceho na polohe oci Avatara a konciaceho
   // vo vrchole maximalneho dostrelu - v body space RB
   a=bricks[i].rb->vertex_to_bs(EP);
   b=bricks[i].rb->vertex_to_bs(TP);
   // INFO: zaciatok a koniec luca - v body space RB
   //printf("a: %.2f %.2f %.2f\n",a.x,a.y,a.z);
   //printf("b: %.2f %.2f %.2f\n",b.x,b.y,b.z);
   // pre vsetky trojholniky, z ktorych je zlozene RB
   loopk(bricks[i].rb->tris_num) {
    t=bricks[i].rb->get_tri(k); // konkretny trojuholnik
    // detekcia kolizie luca a trojuholnika
    if (m3d.ray_tri_collision(a,b,t->n,t->d,t->cn[0],t->cn[1],t->cn[2])) {
     ip=m3d.get_ipoint(); // priesecnik luca a trojuholnika
     d=(ip-a).magnitude();
     if (d<d_min) {       // vyber najblizsieho priesecnika
      ip_min=ip; d_min=d; hit_RB=i; hit_t=k;
     }
    }
   } // pre vsetky trojholniky, z ktorych je zlozene RB
  } // pre vsetky RB

  // ak Avatar zasiahol nejaky trojuholnik
  if (hit_RB!=-1) {
   // INFO: zasiahnute RB a zasiahnuty trojuholnik
   //printf("zasah: %i %i\n",hit_RB,hit_t);
   // zasiahnuty trojuholnik
   t=bricks[hit_RB].rb->get_tri(hit_t);
   // relativna poloha priesecnika vo vztahu k v0 zasiahnuteho trojuholnika
   ip_min_bs=ip_min; ip_min_bs-=t->v0;
   ip_min_bs.x=my_abs(ip_min_bs.x);
   ip_min_bs.y=my_abs(ip_min_bs.y);
   ip_min_bs.z=my_abs(ip_min_bs.z);
   if (ip_min_bs.x<0.0001f) { // rovina Z Y
    u=ip_min_bs.z*1.0f/bricks[hit_RB].rb->sz;
    v=ip_min_bs.y*1.0f/bricks[hit_RB].rb->sy;
   } else if (ip_min_bs.y<0.0001f) { // rovina X Z
           u=ip_min_bs.x*1.0f/bricks[hit_RB].rb->sx;
           v=ip_min_bs.z*1.0f/bricks[hit_RB].rb->sz;
          }
          else { // rovina X Y
           u=ip_min_bs.x*1.0f/bricks[hit_RB].rb->sx;
           v=ip_min_bs.y*1.0f/bricks[hit_RB].rb->sy;
          }
   u*=128.0f; v*=128.0f;
   switch (hit_t) { // posun na spravne miesto v texture
    case 0:
    case 1:  u+=256; v+=256; break; // vpredu
    case 2:
    case 3:  u+=256; v+=384; break; // hore
    case 4:
    case 5:  u+=0;   v+=256; break; // vzadu
    case 6:
    case 7:  u+=256; v+=128; break; // dole
    case 8:
    case 9:  u+=128; v+=256; break; // vlavo
    case 10:
    case 11: u+=384; v+=256; break; // vpravo
   }
   ui=(unsigned int) u; vi=512-(unsigned int) v;
   // uprava textury - doplnenie dier po gulkach
   glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[100+hit_RB]]);
   glTexSubImage2D(GL_TEXTURE_2D,0,ui-4,vi-2,8,4,GL_RGB,GL_UNSIGNED_BYTE,data);
   glTexSubImage2D(GL_TEXTURE_2D,0,ui-2,vi-4,4,2,GL_RGB,GL_UNSIGNED_BYTE,data);
   glTexSubImage2D(GL_TEXTURE_2D,0,ui-2,vi+2,4,2,GL_RGB,GL_UNSIGNED_BYTE,data);
   glTexSubImage2D(GL_TEXTURE_2D,0,ui-3,vi-3,1,1,GL_RGB,GL_UNSIGNED_BYTE,data);
   glTexSubImage2D(GL_TEXTURE_2D,0,ui+2,vi-3,1,1,GL_RGB,GL_UNSIGNED_BYTE,data);
   glTexSubImage2D(GL_TEXTURE_2D,0,ui+2,vi+2,1,1,GL_RGB,GL_UNSIGNED_BYTE,data);
   glTexSubImage2D(GL_TEXTURE_2D,0,ui-3,vi+2,1,1,GL_RGB,GL_UNSIGNED_BYTE,data);

   // poloha priesecnika vo world space
   ip_min=bricks[hit_RB].rb->vertex_to_ws(ip_min);

   // sila gulky je zavisla na vzdialenosti Avatara od RB
   a=(ip_min-EP); d=a.magnitude(); if (d>0.0f) f_vec/=d;
   // na zasiahnute RB posobi impulz
   bricks[hit_RB].rb->add_J(f_vec*500.0f,ip_min);
  }
 } // ak Avatar striela

 // zrychlenie/spomalenie plynutia casu
 DT*=t_scale;
 // cas, ktory uplynul od posledneho zaznamenaneho casoveho okamihu
 t_next=t_last+DT; if (t_next>0.1f) t_next=0.1f;
 // pocet opakovani vypoctov simulacie
 // t_step=1.0f/100.0f=0.01f potom:
 // ak t_next=0.0f => iter_num=0
 // ak t_next=0.1f => iter_num=10
 iter_num=(int) (t_next/t_step);

 // vypocty simulacie budu vykonane iter_num-krat
 loopk(iter_num) {
  loopi(bricks_num) {
   //bricks[i].rb->add_F(); // pridanie externych sil
   bricks[i].rb->integrate(t_step);
  }
  // detekcia a riesenie kolizii
  col_detection.detect_and_resolve_collisions(bricks,bricks_num,planes,2,t_step);
 }

 // zaznamenanie casoveho okamihu po vykonani x vypoctov simulacie
 // ak iter_num=0 vypocty simulacie sa nevykonaju
 // ak iter_num=x vypocty simulacie sa vykonaju x-krat
 t_last=t_next-iter_num*t_step;

 glPushAttrib(GL_ALL_ATTRIB_BITS);
 glEnable(GL_LIGHTING);

 // vykreslenie tehal
 loopi(bricks_num) {
  // vyber textury
  glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[100+i]]);
  bricks[i].draw();
 }

 glPopAttrib();
}

// sklenena platna ---------------------------------------------------------

// konstruktor *************************************************************
CPhGlass::CPhGlass()
{
 t_scale=1.0f;       // zrychlenie/spomalenie plynutia casu
 t_step=1.0f/100.0f; // trvanie 1 vypoctu simulacie
 // cas, ktory uplynul od posledneho zaznamenaneho casoveho okamihu
 t_next=0.0f;
 // casovy okamih po vykonani x vypoctov simulacie
 t_last=0.0f;
 // pocet opakovani vypoctov simulacie
 iter_num=1;

 // na zaciatku je sklenena platna celistva - z 1 kusu
 parts_num=1; // 1 kus
 loopi(parts_num) {
  // polovica rozmerov RB
  parts[i].halfSize=vec3(0.06f,1.4f,1.6f);
  // inicializacia RB
  parts[i].rb->rb_init(parts[i].halfSize);
  // nastavenie parametrov RB
  parts[i].rb->set_M(100.0f);
  I.set_box_inertia_tensor(parts[i].halfSize,100.0f);
  parts[i].rb->set_Ibs(I);
  // normalizacia R, vypocet t_mat, I_inv - vzdy po zmene R, alebo Ibs
  parts[i].rb->update_params();
  // poloha RB
  parts[i].rb->set_s(-14.0f,1.68f,-19.2f);
  // pociatocne gravitacne zrychlenie zapricinene gravitacnou silou
  parts[i].rb->set_a(vec3(0.0f,-9.81f,0.0f));
  parts[i].rb->set_damping(0.9f,0.9f);
  parts[i].rb->set_friction(0.9f);
 }

 // kolizne roviny
 planes[0].normal=vec3(0.0f,1.0f,0.0f);
 planes[0].dist=0.28f;
 planes[1].normal=vec3(1.0f,0.0f,0.0f);
 planes[1].dist=14.953f;
 planes[2].normal=vec3(0.0f,0.0f,1.0f);
 planes[2].dist=21.134f;
 planes[3].normal=vec3(0.0f,0.0f,-1.0f);
 planes[3].dist=-17.286f;

 // nahratie textury pre sklenenu platnu
 texture.load2D(tid[200],"data/textures/glass.png",T_MIPMAP|T_LINEAR|T_CLAMP);
}

// destruktor **************************************************************
CPhGlass::~CPhGlass()
{

}

// vykreslenie sklenenej platne ********************************************
void CPhGlass::draw(vec3 EP,vec3 TP,unsigned char F,bool CF,float DT)
{
 // nebolo zasiahnute ziadne RB
 hit_RB=-1;

 // ak Avatar striela
 if ((F!=0)&&(CF)) {
  // normalizovany dopredu smerujuci vektor
  f_vec=-TP;
  // vzdialenost, minimalna vzdialenost
  d=0.0f; d_min=wRange;
  // vrchol na pozicii maximalneho dostrelu
  TP=EP+TP*wRange;
  // suradnice Avatara su opacne ako suradnice v scene
  EP=-EP; TP=-TP;

  // najdenie najblizsieho trojuholnika patriaceho RB, ktory bol zasiahnuty gulkou
  // pre vsetky RB
  loopi(parts_num) {
   // zaciatok a koniec luca zacinajuceho na polohe oci Avatara a konciaceho
   // vo vrchole maximalneho dostrelu - v body space RB
   a=parts[i].rb->vertex_to_bs(EP);
   b=parts[i].rb->vertex_to_bs(TP);
   // INFO: zaciatok a koniec luca - v body space RB
   //printf("a: %.2f %.2f %.2f\n",a.x,a.y,a.z);
   //printf("b: %.2f %.2f %.2f\n",b.x,b.y,b.z);
   // pre vsetky trojholniky, z ktorych je zlozene RB
   loopk(parts[i].rb->tris_num) {
    t=parts[i].rb->get_tri(k); // konkretny trojuholnik
    // detekcia kolizie luca a trojuholnika
    if (m3d.ray_tri_collision(a,b,t->n,t->d,t->cn[0],t->cn[1],t->cn[2])) {
     ip=m3d.get_ipoint(); // priesecnik luca a trojuholnika
     d=(ip-a).magnitude();
     if (d<d_min) {       // vyber najblizsieho priesecnika
      ip_min=ip; d_min=d; hit_RB=i; hit_t=k;
     }
    }
   } // pre vsetky trojholniky, z ktorych je zlozene RB
  } // pre vsetky RB

  // ak Avatar zasiahol nejaky trojuholnik
  if (hit_RB!=-1) {
   // INFO: zasiahnute RB a zasiahnuty trojuholnik
   //printf("zasah: %i %i\n",hit_RB,hit_t);

   // poloha priesecnika v body space
   ip_min_bs=ip_min;
   // poloha priesecnika vo world space
   ip_min=parts[hit_RB].rb->vertex_to_ws(ip_min);
   // sila gulky je zavisla na vzdialenosti Avatara od RB
   a=(ip_min-EP); d=a.magnitude(); if (d>0.0f) f_vec/=d;

   if ((parts_num+3)<=MAX_PARTS) {
    // zasiahnuty trojuholnik
    t=parts[hit_RB].rb->get_tri(hit_t);
    // relativna poloha priesecnika vo vztahu k v0 zasiahnuteho trojuholnika
    ip_min_bs-=t->v0;
    ip_min_bs.x=my_abs(ip_min_bs.x);
    ip_min_bs.y=my_abs(ip_min_bs.y);
    if (ip_min_bs.z<0.0f) zr[0]=1.0f; else zr[0]=-1.0f;
    zr[1]=zr[0]*(-1.0f);
    ip_min_bs.z=my_abs(ip_min_bs.z);
    if (ip_min_bs.x<0.0001f) { // rovina Z Y
     sx[0]=parts[hit_RB].rb->sx; sy[0]=ip_min_bs.y;                      sz[0]=ip_min_bs.z;
     sx[1]=sx[0];                sy[1]=ip_min_bs.y;                      sz[1]=parts[hit_RB].rb->sz-ip_min_bs.z;
     sx[2]=sx[0];                sy[2]=parts[hit_RB].rb->sy-ip_min_bs.y; sz[2]=parts[hit_RB].rb->sz-ip_min_bs.z;
     sx[3]=sx[0];                sy[3]=parts[hit_RB].rb->sy-ip_min_bs.y; sz[3]=ip_min_bs.z;
     part_pos[0].set(0.0f,-parts[hit_RB].rb->sy/2.0f+sy[0]/2.0f,zr[0]*parts[hit_RB].rb->sz/2.0f+zr[1]*sz[0]/2.0f);
     part_pos[1].set(0.0f,-parts[hit_RB].rb->sy/2.0f+sy[1]/2.0f,zr[1]*parts[hit_RB].rb->sz/2.0f+zr[0]*sz[1]/2.0f);
     part_pos[2].set(0.0f, parts[hit_RB].rb->sy/2.0f-sy[2]/2.0f,zr[1]*parts[hit_RB].rb->sz/2.0f+zr[0]*sz[2]/2.0f);
     part_pos[3].set(0.0f, parts[hit_RB].rb->sy/2.0f-sy[3]/2.0f,zr[0]*parts[hit_RB].rb->sz/2.0f+zr[1]*sz[3]/2.0f);
     loopi(4) part_pos[i]=parts[hit_RB].rb->vertex_to_ws(part_pos[i]);

     // povodne RB zanikne a namiesto neho vznikne 4. kus
     // polovica rozmerov RB
     parts[hit_RB].halfSize=vec3(sx[3]/2.0f,sy[3]/2.0f,sz[3]/2.0f);
     tmp_R=parts[hit_RB].rb->get_R();
     // inicializacia RB
     parts[hit_RB].rb->rb_init(parts[hit_RB].halfSize);
     parts[hit_RB].rb->set_R(tmp_R);
     // nastavenie parametrov RB
     parts[hit_RB].rb->set_M(100.0f);
     I.set_box_inertia_tensor(parts[hit_RB].halfSize,100.0f);
     parts[hit_RB].rb->set_Ibs(I);
     // normalizacia R, vypocet t_mat, I_inv - vzdy po zmene R, alebo Ibs
     parts[hit_RB].rb->update_params();
     // poloha RB
     parts[hit_RB].rb->set_s(part_pos[3]);
     // pociatocne gravitacne zrychlenie zapricinene gravitacnou silou
     parts[hit_RB].rb->set_a(vec3(0.0f,-9.81f,0.0f));
     parts[hit_RB].rb->set_damping(0.9f,0.9f);
     parts[hit_RB].rb->set_friction(0.9f);

     // vzniknu 3 nove kusy
     loopi(3) {
      // polovica rozmerov RB
      parts[parts_num+i].halfSize=vec3(sx[i]/2.0f,sy[i]/2.0f,sz[i]/2.0f);
      // inicializacia RB
      parts[parts_num+i].rb->rb_init(parts[parts_num+i].halfSize);
      parts[parts_num+i].rb->set_R(tmp_R);
      // nastavenie parametrov RB
      parts[parts_num+i].rb->set_M(100.0f);
      I.set_box_inertia_tensor(parts[parts_num+i].halfSize,100.0f);
      parts[parts_num+i].rb->set_Ibs(I);
      // normalizacia R, vypocet t_mat, I_inv - vzdy po zmene R, alebo Ibs
      parts[parts_num+i].rb->update_params();
      // poloha RB
      parts[parts_num+i].rb->set_s(part_pos[i]);
      // pociatocne gravitacne zrychlenie zapricinene gravitacnou silou
      parts[parts_num+i].rb->set_a(vec3(0.0f,-9.81f,0.0f));
      parts[parts_num+i].rb->set_damping(0.9f,0.9f);
      parts[parts_num+i].rb->set_friction(0.9f);

      // na 3 nove kusy posobi impulz
      parts[parts_num+i].rb->add_J(f_vec*500.0f,ip_min);
     }
     parts_num+=3;
    }
   }

   // na zasiahnute RB posobi impulz
   parts[hit_RB].rb->add_J(f_vec*500.0f,ip_min);
  } // ak Avatar zasiahol nejaky trojuholnik
 } // ak Avatar striela

 // zrychlenie/spomalenie plynutia casu
 DT*=t_scale;
 // cas, ktory uplynul od posledneho zaznamenaneho casoveho okamihu
 t_next=t_last+DT; if (t_next>0.1f) t_next=0.1f;
 // pocet opakovani vypoctov simulacie
 // t_step=1.0f/100.0f=0.01f potom:
 // ak t_next=0.0f => iter_num=0
 // ak t_next=0.1f => iter_num=10
 iter_num=(int) (t_next/t_step);

 // vypocty simulacie budu vykonane iter_num-krat
 loopk(iter_num) {
  loopi(parts_num) {
   //parts[i].rb->add_F(); // pridanie externych sil
   parts[i].rb->integrate(t_step);
  }
  // detekcia a riesenie kolizii
  col_detection.detect_and_resolve_collisions(parts,parts_num,planes,4,t_step);
 }

 // zaznamenanie casoveho okamihu po vykonani x vypoctov simulacie
 // ak iter_num=0 vypocty simulacie sa nevykonaju
 // ak iter_num=x vypocty simulacie sa vykonaju x-krat
 t_last=t_next-iter_num*t_step;

 glPushAttrib(GL_ALL_ATTRIB_BITS);
 glDisable(GL_LIGHTING);

 glEnable(GL_BLEND);
 glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);

 // vykreslenie vsetkych kusov sklenenej platne
 loopi(parts_num) {
  // vyber textury
  glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[100]]);
  parts[i].draw();
 }

 glDisable(GL_BLEND);

 glPopAttrib();
}
